"""
Public methods for writing I/Q recordings according to ITU-R SM.2117-0.
"""

import h5py
import numpy as np

from itusm2117._internal import *
from itusm2117._schema import validator


def write_iq_dataset(filename, recordings, sampling_frequency=None, metadata = {}, \
    group = None, dataset_name = None, channel_suffixes = None, mode = "a", **kw):
    """ Method for storing I/Q recordings according to ITU-R SM.2117-0.

    This method is flexibe to support different use cases althouth it currently
    lacks the following ITU-R SM.2117-0 features:
        - BitField datasets
        - Multisector recordings
        - Integer I/Q recordings (H5T_STD_I32LE & H5T_STD_I16LE)

    Example usage:
    >>> from kb_imex.ituh5 import write_iq_dataset
    >>> iq = [1+1j, 2+2j, 3+3j, 4+4j]
    >>> sampling_frequency = 125e5
    >>> write_iq_dataset("my_iq_data.h5", iq, sampling_frequency)
    
    Args:
        filename (str): The filename of the .h5 file where the recordings 
            should be stored. If the file doens't exists it will be created,
            if it does, the recordings will be appended.
        recordings (np.array): The I/Q data itself as a Numpy array. It can be
            a single channel or multiple channels/recordings. If n is the 
            number of time samples, the channel (or channels) can be a complex
            array of shape (n,) or it can have I/Q as a dimension, either being 
            the first dimension (second dimension for multichannel), i.e., one 
            array of shape (2, n) with I samples first Q samples second, or
            being the second dimension (third dimension for multichannel), i.e.
            , each timestep is an I/Q pair of shape (n, 2). Refer to the Enum 
            class RecordingFormat to a completed description of the supported 
            shapes. If a unsupported shape is provided, a ValueError exception
            will be raised.
        sampling_frequency (float): Defaults to None. This is the only 
            obrigatory user provided metadata of the ITU recommedation. 
            Alternatively it can be provided via the metadata dictionary (using 
            sampling_frequency as key or "Sampling frequency (Hz)"). If the 
            sampling frequency is not provided via this paramater or via the 
            dictionary, a ValueError exception will be raised.
        metadata (dict, optional): Defaults to empty dict. A dictionary with
            the metadata associated with the recordings/channels. The keys of
            the dictionary can either be the standard strings used in the
            ITU recommendation or underscored shorthands, or a combination. If
            shorthands are used, they can also be passed as standard method
            parameters. Bellow is a complete table of all metadata, with 
            standard ITU strings and shorthands. For a complete description of
            each metadata/attribute and it's accepted values, refer to the ITU 
            SM.2117-0 recommendation. Note that all metadata values are 
            validated according to the ITU rules (e.g., types, accepted 
            values). Fields not specified in the Recommendation are preffixed 
            with 'User ' unless they already starts with 'User'.
            | KB IMEX Column            | ITU Column                          |
            | _________________________ |____________________________________ |
            | (readonly)                | ITU_R data set class                |
            | (readonly)                | ITU_R Recommendation                |
            | carrier_frequency         | RF carrier frequency (Hz)           |
            | sampling_frequency        | Sampling frequency (Hz)             |
            | (readonly)                | Data set type interpretation        |
            | unit                      | Data set unit                       |
            | scaling                   | Data set scaling factor             |
            | comment                   | Comment                             |
            | device                    | Device                              |
            | bandwidth                 | Filter bandwidth (Hz)               |
            | timestamp                 | Timestamp coarse (s)                |
            | timestamp_fine            | Timestamp fine (ns)                 |
            | latitude                  | Geolocation latitude (degree)       |
            | longitude                 | Geolocation longitude (degree)      |
            | altitude                  | Geolocation altitude (m)            |
            | separation                | Geolocation separation (m)          |
            | speed                     | Speed over ground magnitude (m/s)   |
            | speed_azimuth             | Speed over ground azimuth (degree)  |
            | orientation_azimuth       | Orientation azimuth (degree)        |
            | orientation_elevation     | Orientation elevation (degree)      |
            | orientation_skew          | Orientation skew (degree)           |
            | magnetic_declination      | Magnetic declination (degree)       |
            | unsynced_flag             | Unsynced timestamp flag             |
            | invalid_flag              | Invalid flag                        |
            | pllunlocked_flag          | PLL unlocked                        |
            | agc_flag                  | AGC flag                            |
            | detected_flag             | Detected signal flag                |
            | spectral_inversion_flag   | Spectral inversion flag             |
            | overrange_flag            | Over range flag                     |
            | lostsample_flag           | Lost sample flag                    |
            | attenuator                | Attenuator (dB)                     |
            | antenna_factor            | Antenna factor (1/m)                |
            | reference                 | Reference point                     |
            | receiver_impedance        | Receiver input impedance (Ohm)      |
        group (str, list or tuple): Where to place the dataset. Defaults to 
            None. If not provided, the dataset will be placed at the root of 
            the .h5 file. If it's a string, the dataset will be placed under
            the group with specified name, and if a list or tuple of strings is
            provide, the sequence will be used as subgroups. If the any group
            doesn't exists it is created.
        dataset_name (str): The name of the dataset. Defaults to None. If not 
            provided, an auto generated name will be given with the preffix 
            Dataset_X, with X being an integer auto incremented. If a 
            duplicated name is given an error is raised.
        channel_suffixes (str, list or tuple): Name suffixes to be used for 
            each channel/recording. If not provided, a zero-based integer
            index is used.
        mode (str): Writing mode. "a" (default) for appending a new dataset and 
            "w" for creating a new file. Both modes creates a new file if it 
            doesn't exits.
        **kw: Additional metadata.

    Return:
        None
    """
    
    # Initialization
    metadata.update(kw)
    if sampling_frequency is not None:
        metadata["sampling_frequency"] = sampling_frequency    
    if isinstance(recordings, list):
        recordings = np.array(recordings)
    if not filename.endswith(".h5"):
        filename = filename + ".h5"

    # Format validation
    format = determine_format(recordings)
    if format is RecordingFormat.INVALID:
        raise ValueError(f"Invalid recording shape {recordings.shape}")
    if format in MULTIPLE_RECORDINGS and channel_suffixes is not None:
        if len(channel_suffixes) != recordings.shape[0]:
            raise ValueError(f"Number of provided recordings {recordings.shape[0]} differ from \
                number of provided channel suffixes {len(channel_suffixes)}.")

    # Cast data to appropriate type
    if format in [RecordingFormat.MULTIPLE_COMPLEX, RecordingFormat.SINGLE_COMPLEX]:
        if recordings.dtype != np.complex64:
            recordings = recordings.astype(np.complex64)
    else:
        if recordings.dtype != np.float32:
            recordings = recordings.astype(np.float32)


    # Prepare channel suffix generator
    if channel_suffixes is None:
        suffix_generator = index_generator(recordings.shape[0])
        suffix = 0
    else:
        if isinstance(channel_suffixes, str):
            channel_suffixes = [channel_suffixes]              
        suffix_generator = list_generator(channel_suffixes)
        suffix = channel_suffixes[0]

    # Normalize & order metadata
    normalized_metadata = validator.validated(metadata)
    if normalized_metadata is None:
        raise ValueError("\n".join([str(f"{k}: {v}") for k, v in validator.errors.items()]))    
    normalized_metadata = reorder_metadata(normalized_metadata) 

    # Create or navigate group
    h5 = h5py.File(filename, mode)
    if group is None:
        group = h5
    else:
        if isinstance(group, str):
            group = create_or_navigate_groups(h5, [group])
        else:
            group = list(group)
            group.reverse()
            group = create_or_navigate_groups(h5, group) 

    # Define dataset name
    if dataset_name is None:
        dataset_name = f"Dataset_{get_next_dataset_index(group)}"
    else:
        if dataset_name in group:
            raise ValueError("Dataset name {dataset_name} already exists in group.")

    # Store iq data
    if format in MULTIPLE_RECORDINGS:    
        suffixes = list(suffix_generator)
        dataset = _store_recordings(group, dataset_name, recordings, format, suffixes)
    else:    
        dataset = _store_recording(group, dataset_name, recordings, format, suffix)

    # Store metadata & exit
    for k,v in normalized_metadata.items():
        dataset.attrs.create(k, v)
    h5.close()

    return normalized_metadata


def _store_recordings(group, dataset_name, recordings, format, suffixes):
    basic_channel_dtype = [('Real', 'float32'), ('Imag', 'float32')]
    names = [f"Channel_{s}" for s in suffixes]
    types = np.dtype([(name, basic_channel_dtype) for name in names])
    n = format.get_size(recordings)
    dataset = group.create_dataset(dataset_name, (n,), dtype=types)
    parse = format.get_parsing_function()
    for i in range(0, len(names)):
        lst = parse(recordings[i])
        dataset[names[i]] = lst
    return dataset       

def _store_recording(group, dataset_name, recording, format, suffix):
    basic_channel_dtype = np.dtype([('Real', 'float32'), ('Imag', 'float32')])
    name = f"Channel_{suffix}"
    typee = [(name, basic_channel_dtype)]
    n = format.get_size(recording)
    dataset = group.create_dataset(dataset_name, (n,), dtype=typee)
    parse = format.get_parsing_function()
    lst = parse(recording)
    dataset[name] = lst
    # dataset[name]["Real"] = r
    # dataset[name]["Imag"] = i
    return dataset
    

# def _store_recording(group, recording, channel_suffix):
#         channel_key = f"Channel_{channel_suffix}"
#         if len(recording.shape) > 2:
#             raise ValueError(f"IQ recording with unsupported shape of {recording.shape}.")
#         if len(recording.shape) == 2:
#             if recording.shape[0] == 2 and recording.shape[1] !=2:
#                 real = recording[0]
#                 imag = recording[1]
#                 n = recording.shape[1]
#             elif recording.shape[1] == 2 and recording.shape[0] !=2:
#                 real = recording[:,0]
#                 imag = recording[:,1]
#                 n = recording.shape[0]
#             else:
#                 raise ValueError(f"IQ recording with unsupported shape of {recording.shape}.")        
#         elif len(recording.shape) == 1:
#             real = np.real(recording)
#             imag = np.imag(recording)
#             n = recording.shape[0]
#         channel_dataset = group.create_dataset(channel_key, (n,), dtype=np.dtype([('Real', 'float32'), ('Imag', 'float32')]))
#         channel_dataset["Real"] = real
#         channel_dataset["Imag"] = imag


"""
Public methods for reading I/Q recordings according to ITU-R SM.2117-0.
"""

import h5py
import numpy as np

from itusm2117._internal import *

def read_iq_dataset(filename, dataset_name, group = None):
    """ Method for reading I/Q recordings according to ITU-R SM.2117-0.   

    Example usage:
    >>> from kb_imex.ituh5 import write_iq_dataset, read_iq_dataset
    >>> import numpy as np
    >>> iq = [1+1j, 2+2j, 3+3j, 4+4j]
    >>> sampling_frequency = 125e5
    >>> write_iq_dataset("my_iq_data.h5", iq, sampling_frequency)
    >>> metadata, recordings, channels = read_iq_dataset("my_id_data.h5", "Dataset_0")
    >>> assert np.array_equal(recordings[0], np.array(iq))
    True
    >>> assert channels[0] == "Channel_0"
    True
    >>> assert metadata["Sampling frequency (Hz)"] == 125e6
    True

    Args:
        filename (str): Filename of the .h5 file to be read.
        dataset_name (str): Name of the dataset to be read.
        group (str, list or tuple): H5 group where the dataset resides. If a 
            list or tuple is provided, each element is considered a subgroup
            withing a tree structure. Defaults to None (root).        
    
    """

    h5 = h5py.File(filename, "r")
    try:
        if group is None:
            group = h5
        elif isinstance(group, str):
            group = h5[group]
        else:
            group = list(group)
            group.reverse()
            group = navigate_groups(h5, group)

        dataset = group[dataset_name]
        channels = dataset.dtype.names

        results = []
        for channel in channels:
            re =  dataset[channel]["Real"]
            im =  dataset[channel]["Imag"]
            results.append(re + (im * 1j))
        results = np.stack(results)

        metadata = {}
        for k, v in dataset.attrs.items():
            metadata[k] = v

        return metadata, results, channels
    
    finally:
        h5.close()

    




    
    
    


"""Private helper methods of the kb_imex.ituh5 package."""

from enum import Enum

import numpy as np

from itusm2117._schema import ORDER

def navigate_groups(parent, groups):
    if len(groups) == 1:
        group = groups[0]
        if group in parent:
            return parent[group]
        else:
            raise IndexError("Group name not found")
    else:
        group = groups.pop()
        if group in parent:
            parent = parent[group]
        else:
            raise IndexError("Group name not found")
        return create_or_navigate_groups(parent, groups)

def create_or_navigate_groups(parent, groups):
    if len(groups) == 1:
        group = groups[0]
        if group in parent:
            return parent[group]
        else:
            return parent.create_group(group)
    else:
        group = groups.pop()
        if group in parent:
            parent = parent[group]
        else:
            parent = parent.create_group(group)
        return create_or_navigate_groups(parent, groups)

def index_generator(size):
    for i in range(0, size):
        yield i

def get_next_dataset_index(group):
    keys = group.keys()
    suffixes = ["_".join(key.split("_")[1:]) for key in keys if key.startswith("Dataset_")]
    numbers = [int(s) for s in suffixes if s.isdigit()]
    numbers.sort()
    if len(numbers) == 0:
        next_int = 0
    else:
        next_int = numbers[len(numbers)-1] + 1
    return next_int

def list_generator(suffixes):
    for suffix in suffixes:
        yield suffix        

def reorder_metadata(metadata):
        result = {}        
        for key in ORDER:
            if key in metadata:
                result[key] = metadata[key]
        for k, v in metadata.items():
            if k not in result:
                if k.startswith("User"):
                    result[k] = v
                else:
                    result["User " + k] = v
        return result

class RecordingFormat(Enum):
    SINGLE_COMPLEX = (None,)
    SINGLE_IQ_FIRST = (2, None)
    SINGLE_TIME_FIRST = (None, 2)
    MULTIPLE_COMPLEX = (None, None)
    MULTIPLE_IQ_FIRST = (None, 2, None)
    MULTIPLE_TIME_FIRST = (None, None, 2)
    INVALID = 6

    def get_size(self, recordings):
        if self ==  RecordingFormat.SINGLE_COMPLEX:
            return recordings.shape[0]
        elif self == RecordingFormat.SINGLE_IQ_FIRST:
            return recordings.shape[1]
        elif self == RecordingFormat.SINGLE_TIME_FIRST:
            return recordings.shape[0]
        elif self ==  RecordingFormat.MULTIPLE_COMPLEX:
            return recordings.shape[1]
        elif self == RecordingFormat.MULTIPLE_IQ_FIRST:
            return recordings.shape[2]
        elif self == RecordingFormat.MULTIPLE_TIME_FIRST:
            return recordings.shape[1]

    def get_parsing_function(self):
        def complex_parse(recording):
            # if recording.dtype != np.complex64:
            #     recording = np.complex64(recording)
            real = np.real(recording)
            imag = np.imag(recording)
            arr = np.float32([real, imag])
            newarr = np.swapaxes(arr,0,1)
            return list(map(tuple, newarr))
            # newarr = newarr.view(dtype = np.dtype([("Real", newarr.dtype), ("Imag", newarr.dtype)]))
            # newarr = newarr.reshape(arr.shape[:-1])
            # return arr   

        def iqfirst_parse(recording):
            newarr = np.swapaxes(np.float32(recording),0,1)
            return list(map(tuple, newarr))
            # if recording.dtype != np.float32:
            #     recording = np.float32(recording)
            # real = recording[0]
            # imag = recording[1]
            # return real, imag
        def timefirst_parse(recording):
            newarr = np.float32(recording)
            return list(map(tuple, newarr))
            # if recording.dtype != np.float32:
            #     recording = np.float32(recording)
            # real = recording[:,0]
            # imag = recording[:,1]
            # return real, imag

        if (self == RecordingFormat.SINGLE_COMPLEX 
            or self == RecordingFormat.MULTIPLE_COMPLEX):
            return complex_parse
        elif (self == RecordingFormat.SINGLE_IQ_FIRST
            or self == RecordingFormat.MULTIPLE_IQ_FIRST):
            return iqfirst_parse
        else:
            return timefirst_parse

MULTIPLE_RECORDINGS = set((RecordingFormat.MULTIPLE_COMPLEX, RecordingFormat.MULTIPLE_IQ_FIRST, RecordingFormat.MULTIPLE_TIME_FIRST))

SINGLE_RECORDING = set((RecordingFormat.SINGLE_COMPLEX, RecordingFormat.SINGLE_IQ_FIRST, RecordingFormat.SINGLE_TIME_FIRST))

def determine_format(recordings):

    if len(recordings.shape) == 1:
        return RecordingFormat.SINGLE_COMPLEX
    
    elif len(recordings.shape) == 2:
        if recordings.shape[0] == 2 and recordings.shape[1] !=2 and not np.iscomplexobj(recordings):
            return RecordingFormat.SINGLE_IQ_FIRST
        elif recordings.shape[1] == 2 and recordings.shape[0] !=2:
            return RecordingFormat.SINGLE_TIME_FIRST
        else:
            return RecordingFormat.MULTIPLE_COMPLEX
    
    elif len(recordings.shape) == 3:
        if recordings.shape[1] == 2 and recordings.shape[2] !=2:
            return RecordingFormat.MULTIPLE_IQ_FIRST
        elif recordings.shape[2] == 2 and recordings.shape[1] !=2:
            return RecordingFormat.MULTIPLE_TIME_FIRST
        else:
            return RecordingFormat.INVALID

    else:
        return RecordingFormat.INVALID


"""Cerberus Schema definition and related methods."""

from dateutil import parser
from datetime import datetime

from cerberus import Validator

def _to_utc_fine(value):
    if isinstance(value, datetime):
        return value.timestamp()
    elif isinstance(value, str):
        dt = parser.parse(value)
        return dt.timestamp()
    elif isinstance(value, float):
        return value    
    raise ValueError("Unsupported date representation.")

def _to_utc(value):
    if isinstance(value, datetime):
        return int(value.timestamp())
    elif isinstance(value, str):
        dt = parser.parse(value)
        return int(dt.timestamp())
    elif isinstance(value, float):
        return int(value)
    elif isinstance(value, int):
        return value    
    raise ValueError("Unsupported date representation.")

SCHEMA = {    
    "ITU-R data set class":{
        "required": True,
        "type": "string",
        "readonly": True, 
        "default": "I/Q",
    },
    "ITU-R Recommendation":{
        "required": True,
        "type": "string",
        "readonly": True, 
        "default": "Rec. ITU-R SM.2117-0",        
    },
    "carrier_frequency":{
        "rename": "RF carrier frequency (Hz)"
    },
    "RF carrier frequency (Hz)":{
        "required": True,
        "type": "float",
        "min": 0,
        "default": 0,
    },
    "sampling_frequency":{
        "rename": "Sampling frequency (Hz)"
    },
    "Sampling frequency (Hz)":{
        "required": True,
        "type": "float",
        "forbidden": [0],
        "min": 0,
        "rename": "Sampling frequency (Hz)"
    },
    "Data set type interpretation":{
        "required": True,
        "type": "string",
        "readonly": True, 
        "default": "Integer types, used to store I/Q data, are interpreted as fix point numbers with the radix point right to the most significant bit.",    
    },
    "unit":{
        "rename": "Data set unit"
    },
    "Data set unit":{
        "required": True,
        "type": "string",        
        "default": "",
        "allowed": ["", "V", "V/m", "A/m"],        
    },
    "scaling":{    
        "rename": "Data set scaling factor"
    },
    "Data set scaling factor":{
        "required": True,
        "type": "float",        
        "default": 1,        
        "rename": "Data set scaling factor"
    },
    "device":{
        "rename": "Device"
    },
    "Device": {
        "type": "string"
    },
    "comment":{        
        "rename": "Comment"
    },
    "Comment":{
        "type": "string"        
    },
    #need to by lower than sampling_frequency
    "bandwidth":{
        "rename": "Filter bandwidth (Hz)"
    },
    "Filter bandwidth (Hz)":{
        "type": "float"
    },
    "timestamp":{
        "rename": "Timestamp coarse (s)"
    },
    "Timestamp coarse (s)":{
        "type": "integer",
        "coerce": _to_utc,
        "min": 0,
        "max": 1767150000
    },
    "timestamp_fine":{
         "rename": "Timestamp fine (ns)"
    },
    "Timestamp fine (ns)":{
        "type": "float",
        "coerce": _to_utc_fine,
        "min": 0,
        "max": 1767150000        
    },
    "latitude":{
         "rename": "Geolocation latitude (degree)"
    },
    "Geolocation latitude (degree)":{
        "type": "float",        
        "min": -180,
        "max": 180
    },
    "longitude":{
        "rename": "Geolocation longitude (degree)"
    },
    "Geolocation longitude (degree)":{
        "type": "float",        
        "min": -90,
        "max": 80
    },
    "altitude":{       
        "rename": "Geolocation altitude (m)"
    },
    "Geolocation altitude (m)":{
        "type": "float",        
        "min": -10e3
    },
    "separation":{                
        "rename": "Geolocation separation (m)"
    },
    "Geolocation separation (m)":{
        "type": "float"
    },
    "speed":{       
        "rename": "Speed over ground magnitude (m/s)"
    },
    "Speed over ground magnitude (m/s)":{
        "type": "float",        
        "min": 0
    },
    "speed_azimuth":{
        "rename": "Speed over ground azimuth (degree)"
    },
    "Speed over ground azimuth (degree)":{
        "type": "float",        
        "min": 0,
        "max": 360
    },
    "orientation_azimuth":{
        "rename": "Orientation azimuth (degree)"
    },
    "Orientation azimuth (degree)":{
        "type": "float",        
        "min": 0,
        "max": 360
    },
    "orientation_elevation":{
        "rename": "Orientation elevation (degree)"
    },
    "Orientation elevation (degree)":{
        "type": "float",        
        "min": -90,
        "max": 90
    },
    "orientation_skew":{
        "rename": "Orientation skew (degree)"
    },
    "Orientation skew (degree)":{
        "type": "float",        
        "min": -180,
        "max": 180
    },
    ######
    "magnetic_declination":{
        "rename": "Magnetic declination (degree)"
    },
    "unsynced_flag":{   
        "rename": "Unsynced timestamp flag"
    },
    "invalid_flag":{ 
        "rename": "Invalid flag"
    },
    "pllunlocked_flag":{  
        "rename": "PLL unlocked"
    },
    "agc_flag":{    
        "rename": "AGC flag"
    },
    "detected_flag":{   
        "rename": "Detected signal flag"
    },
    "spectral_inversion_flag":{        
        "rename": "Spectral inversion flag"
    },
    "overrange_flag":{    
        "rename": "Over range flag"
    },
    "lostsample_flag":{   
        "rename": "Lost sample flag"
    },
    "attenuator":{                 
        "rename": "Attenuator (dB)"
    },
    "antenna_factor":{
        "rename": "Antenna factor (1/m)"
    },
    "reference":{
        "rename": "Reference point"    
    },
    "receiver_impedance":{
        "rename": "Receiver input impedance (Ohm)"
    },
    "Magnetic declination (degree)":{
        "type": "float",        
        "min": 0,
        "max": 360,
        "dependencies": "Orientation azimuth (degree)"        
    },
    "Unsynced timestamp flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "Invalid flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "PLL unlocked":{
        "type": "integer",        
        "min": 0,                       
    },
    "AGC flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "Detected signal flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "Spectral inversion flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "Over range flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "Lost sample flag":{
        "type": "integer",        
        "min": 0,                       
    },
    "Attenuator (dB)":{
        "type": "float",                          
    },
    "Antenna factor (1/m)":{
        "type": "float"        
    },
    "Reference point":{
        "type": "string",
        "allowed": ["Antenna output port", "Receiver output port"]        
    },
    "Receiver input impedance (Ohm)":{
        "type": "float"        
    }
}

ORDER = ["ITU-R data set class",
"ITU-R Recommendation",
"RF carrier frequency (Hz)",
"Sampling frequency (Hz)",
"Data set type interpretation",
"Data set unit",
"Data set scaling factor",
"Comment",
"Device",
"Filter bandwidth (Hz)",
"Timestamp coarse (s)",
"Timestamp fine (ns)",
"Geolocation latitude (degree)",
"Geolocation longitude (degree)",
"Geolocation altitude (m)",
"Geolocation separation (m)",
"Speed over ground magnitude (m/s)",
"Speed over ground azimuth (degree)",
"Orientation azimuth (degree)",
"Orientation elevation (degree)",
"Orientation skew (degree)",
"Magnetic declination (degree)",
"Unsynced timestamp flag",
"Invalid flag",
"PLL unlocked",
"AGC flag",
"Detected signal flag",
"Spectral inversion flag",
"Over range flag",
"Lost sample flag",
"Attenuator (dB)",
"Antenna factor (1/m)",
"Reference point",
"Receiver input impedance (Ohm)"]

def _user_attrs(field):
    if field.startswith("User"):
        return field
    else:
        return "User " + str(field)

validator = Validator(SCHEMA, allow_unknown={'rename_handler': _user_attrs})
